from copy import deepcopy
import MDAnalysis as mda
import mdtraj as md
from openmm import *
from openmm.app import *
from openmm.unit import *

class PairwiseInteractionEnergy:
    def __init__(self, topology: str, trajectory: str, sel1: str, sel2: str):
        if 'prmtop' in topology:
            self.top = AmberPrmtopFile(topology)
            self.system = self.top.createSystem(nonbondedMethod=CutoffNonPeriodic,
                                                nonbondedCutoff=2. * nanometers,
                                                constraints=HBonds)
        elif 'pdb' in topology:
            self.top = PDBFile(topology)
            forcefield = ForceField('amber14-all.xml', 'implicit/gbn2.xml')
            self.system = forcefield.createSystem(self.top,
                                                  soluteDielectric=1.,
                                                  solventDielectric=80.)
        else:
            raise ValueError('Need prmtop or pdb for topology!')

        self.traj = trajectory
        self.integrator = LangevinMiddleIntegrator(
            300*kelvin, 
            1/picosecond, 
            0.004*picoseconds
        )
        
        idx1 = ()
        idx2 = ()
        self.sels = [idx1, idx2, idx1 + idx2]

    def subset_traj(self, sub_ind: list[str]):
        topology = mm.Topology.from_openmm(self.top)
        sub_top = topology.subset(sub_ind)
        new_top = sub_top.to_openmm()

        structure = pmd.openmm.load_topology(self.top, self.system)[sub_ind]
        new_system = structure.createSystem(soluteDielectric=1.,
                                            solventDielectric=80.)
        simulation = Simulation(new_top, new_system, self.integrator)
        return simulation

    def initialize_systems(self):
        simulations = [self.subset_traj(sel) for sel in self.sels]
        simulations.append(deepcopy(simulations[-1]))
        self.simulations = simulations

    def compute_energy(self):
        full_traj = md.load(self.traj, top=self.top)
        self.energies = np.zeros((full_traj.n_frames, 4))
        for i, sim in enumerate(self.simulations):
            for fr in range(full_traj.n_frames):
                try:
                    sel = self.sels[i]
                except IndexError:
                    sel = self.sels[-1]

                coords = full_traj.xyz[fr, sel, :]
                self.energies[fr, i] = self.calc_energy(sim, coords)
    
    def calc_energy(self, 
                    simulation: Simulation, 
                    positions: np.ndarray) -> float:
        simulation.context.setPositions(positions)
        state = simulation.context.getState(getEnergy=True)
        return state.getPotentialEnergy()._value

    def split_off_components(self):
        total = self.energies[:, 3] - np.sum(self.energies[:, :2], axis=0)
        self.lj = self.energies[:, 4]
        self.el = total - self.lj



#!/usr/bin/env python
import gc
from glob import glob
from MDAnalysis.lib.util import convert_aa_code
import os
from pathlib import Path
from pdbfixer import PDBFixer
from pdbfixer.pdbfixer import Sequence
from rdkit import Chem
import shutil
from typing import Dict, List, Union

PathLike = Union[str, Path]
OptPath = Union[str, Path, None]
Sequences = List[Sequence]

class LigandError(Exception):
    def __init__(self, message='This system contains ligands which we cannot model!'):
        self.message = message
        super.__init__(self.message)

class ImplicitSolvent:
    """
    Class for building a system using ambertools. Produces explicit solvent cubic box
    with user-specified padding which has been neutralized and ionized with 150mM NaCl.
    """
    def __init__(self, path: PathLike, pdb: str, out: OptPath=None, 
                 protein: bool=True, rna: bool=False, dna: bool=False,
                 phos_protein: bool=False, mod_protein: bool=False):
        self.path = path
        os.makedirs(str(path), exist_ok=True)

        self.pdb = pdb

        if out is not None:
            self.out = out
        else:
            self.out = path

        os.makedirs(str(self.out), exist_ok=True)

        switches = [protein, rna, dna, phos_protein, mod_protein]
        ffs = [
            'leaprc.protein.ff19SB', 
            'leaprc.RNA.Shaw', 
            'leaprc.DNA.OL21',
            'leaprc.phosaa19SB',
            'leaprc.protein.ff14SB_modAA'
        ]

        self.ffs = [
            ff for ff, switch in zip(ffs, switches) if switch
        ]
    
    def build(self) -> None:
        """
        Orchestrate the various things that need to happen in order
        to produce an explicit solvent system. This includes running
        `pdb4amber`, computing the periodic box size, number of ions
        needed and running `tleap` to make the final system.
        """
        self.pdbfixit()

    def pdbfixit(self, add_hydrogen=True) -> str:
        fixer = PDBFixer(filename=str(self.path / self.pdb))
        fixer.findMissingResidues()
        fixer.findMissingAtoms()
        fixer.addMissingAtoms()
        fixer.addMissingHydrogens()
        
        out = self.out / 'protein.pdb'
        with open(str(out), 'w') as f:
            PDBFile.writeFile(fixer.topology, fixer.positions, f)

        del fixer
        gc.collect()
    
class ExplicitSolvent(ImplicitSolvent):
    """
    Class for building a system using ambertools. Produces explicit solvent cubic box
    with user-specified padding which has been neutralized and ionized with 150mM NaCl.
    """
    def __init__(self, path: str, pdb: str, padding: float=10., protein: bool=True,
                 rna: bool=False, dna: bool=False, polarizable: bool=False):
        super().__init__(path, pdb, protein, rna, dna)
        
        self.out = f'{path}/system'
        self.pad = padding
        self.ffs.extend(['leaprc.water.opc'])
        self.water_box = 'OPCBOX'
        
        if polarizable:
            self.ffs[0] = 'leaprc.protein.ff15ipq'
            self.ffs[-1] = 'leaprc.water.spceb'
            self.water_box = 'SPCBOX'
    
    def build(self) -> None:
        """
        Orchestrate the various things that need to happen in order
        to produce an explicit solvent system. This includes running
        `pdb4amber`, computing the periodic box size, number of ions
        needed and running `tleap` to make the final system.
        """
        self.prep_pdb()
        dim = self.get_pdb_extent()
        num_ions = self.get_ion_numbers(dim**3)
        self.assemble_system(dim, num_ions)
        self.clean_up_directory()

    def prep_pdb(self):
        os.system(f'pdb4amber -i {self.pdb} -o {self.path}/protein.pdb -y')
        self.pdb = f'{self.path}/protein.pdb'
        
    def assemble_system(self, dim: float, num_ions: int) -> None:
        """
        Build system in tleap.
        """
        tleap_ffs = '\n'.join([f'source {ff}' for ff in self.ffs])
        tleap_complex = f"""{tleap_ffs}
        PROT = loadpdb {self.pdb}
        
        setbox PROT centers
        set PROT box {{{dim} {dim} {dim}}}
        solvatebox PROT {self.water_box} {{0 0 0}}
        
        addions PROT Na+ 0
        addions PROT Cl- 0
        
        addIonsRand PROT Na+ {num_ions} Cl- {num_ions}
        
        saveamberparm PROT {self.out}.prmtop {self.out}.inpcrd
        quit
        """
        
        leap_file = self.write_leap(tleap_complex)
        tleap = f'tleap -f {leap_file}'
        os.system(tleap)
        
    def write_leap(self, inp: str) -> str:
        """
        Writes out a tleap input file and returns the path
        to the file.
        """
        leap_file = f'{self.path}/tleap.in'
        with open(leap_file, 'w') as outfile:
            outfile.write(inp)
            
        return leap_file

    def get_pdb_extent(self) -> int:
        """
        Identifies the longest axis of the protein in terms of X/Y/Z
        projection. Not super accurate but likely good enough for determining
        PBC box size. Returns longest axis length + 2 times the padding
        to account for +/- padding.
        """
        lines = [line for line in open(self.pdb).readlines() if 'ATOM' in line]
        xs, ys, zs = [], [], []
        
        for line in lines:
            x, y, z = [float(i.strip()) for i in line[26:54].split()]
            xs.append(x)
            ys.append(y)
            zs.append(z)
        
        xtent = (max(xs) - min(xs))
        ytent = (max(ys) - min(ys))
        ztent = (max(zs) - min(zs))
        
        return int(max([xtent, ytent, ztent]) + 2 * self.pad)
    
    def clean_up_directory(self) -> None:
        """
        Remove leap log. This is placed wherever the script calling it
        runs and likely will throw errors if multiple systems are
        being iteratively built.
        """
        os.remove('leap.log')
        os.makedirs(f'{self.path}/build', exist_ok=True)
        for f in glob(f'{self.path}/*'):
            if not any([ext in os.path.basename(f) for ext in ['.prmtop', '.inpcrd', 'build']]):
                path, name = os.path.split(f)
                shutil.move(f, f'{path}/build/{name}')
        
    @staticmethod
    def get_ion_numbers(volume: int) -> float:
        """
        Returns the number of Chloride? ions required to achieve 150mM
        concentration for a given volume. The number of Sodium counter
        ions should be equivalent.
        """
        return round(volume * 10e-6 * 9.03)
        
class PLINDERBuilder(ImplicitSolvent):
    """
    Builds complexes consisting of a biomolecule pdb and small molecule ligand.
    Runs antechamber workflow to generate gaff2 parameters.
    """
    def __init__(self, 
                 path: PathLike, 
                 system_id: str,
                 out: PathLike, 
                 **kwargs):
        super().__init__(path / system_id, 'receptor.pdb', out / system_id, 
                         protein=True, rna=True, dna=True, phos_protein=True,
                         mod_protein=True, **kwargs)
        self.system_id = system_id
        self.ffs.append('leaprc.gaff2')
        self.build_dir = self.out / 'build'
        self.ions = None
    
    def build(self) -> None:
        ligs = self.migrate_files()

        if not ligs:
            raise LigandError

        self.ligs = self.ligand_handler(ligs)
        self.assemble_system()

    def ligand_handler(self, ligs: List[PathLike]) -> List[PathLike]:
        ligands = []
        for i, lig in enumerate(ligs):
            lig_builder = LigandBuilder(self.build_dir, lig, i)
            lig_builder.parameterize_ligand()
            ligands.append(os.path.basename(lig)[:-4])

        return ligands
        
    def migrate_files(self) -> List[str]:
        os.makedirs(str(self.build_dir), exist_ok=True)

        # grab the sequence file to complete protein modeling
        shutil.copy(str(self.path / 'sequences.fasta'),
                    str(self.build_dir))
        self.fasta = str(self.build_dir / 'sequences.fasta')
        
        # fix and move pdb
        self.prep_protein()

        # move ligand(s)
        ligands = []
        lig_files = self.path / 'ligand_files'
        ligs = [Path(lig) for lig in glob(str(lig_files) + '/*.sdf')]
        for lig in ligs:
            shutil.copy(str(lig), 
                        str(self.build_dir))

            if self.check_ligand(lig):
                ligands.append(lig.name)


        # handle any potential ions
        if self.ions is not None:
            self.ffs.append('leaprc.water.tip3p')
            self.place_ions()

        return ligands

    def place_ions(self) -> None:
        """
        This is horrible and I apologize profusely if you find yourself
        having to go through the following. Good luck.
        """
        pdb_lines = open(self.pdb).readlines()[:-1]

        if 'TER' in pdb_lines[-1]:
            ln = -2
        else:
            ln = -1

        next_atom_num = int(pdb_lines[ln][6:12].strip()) + 1
        next_resid = int(pdb_lines[ln][22:26].strip()) + 1
        
        for ion in self.ions:
            for atom in ion:
                # HERE BE DRAGONS
                ion_line = f'ATOM  {next_atom_num:>5}'

                if atom[0].lower() in ['na', 'k', 'cl']:
                    ionname = atom[0] + atom[1]
                    ion_line += f'{ionname:>4}  '
                else:
                    ionname = atom[0].upper()
                    ion_line += f'{ionname:>3}   '

                ion_line += f'{ionname:<3}'

                coords = ''.join([f'{x:>8.3f}' for x in atom[2:]])
                ion_line += f'{next_resid:>5}    {coords}  0.00  0.00\n'

                pdb_lines.append(ion_line)
                pdb_lines.append('TER\n')

                next_atom_num += 1
                next_resid += 1

        pdb_lines.append('END')
        
        with open(self.pdb, 'w') as f:
            f.write(''.join(pdb_lines))

    def assemble_system(self) -> None:
        """
        Slightly modified from the parent class, now we have to add
        the ligand parameters and assemble a complex rather than just
        placing a biomolecule in the water box.
        """
        tleap_complex = [f'source {ff}' for ff in self.ffs]
        structs = [f'PROT = loadpdb {self.pdb}']
        combine = 'COMPLEX = combine{PROT'
        for i, lig in enumerate(self.ligs):
            ligand = self.build_dir / lig
            tleap_complex += [f'loadamberparams {ligand}.frcmod', 
                              f'loadoff {ligand}.lib']
            structs += [f'LG{i} = loadmol2 {ligand}.mol2']
            combine += f' LG{i}'
        
        combine += '}'
        tleap_complex += structs
        tleap_complex.append(combine)
        tleap_complex += [
            'set default PBRadii mbondi3',
            f'savepdb COMPLEX {self.out}/system.pdb',
            f'saveamberparm COMPLEX {self.out}/system.prmtop {self.out}/system.inpcrd',
            'quit'
        ]

        tleap_complex = '\n'.join(tleap_complex)
        leap_file = self.build_dir / 'tleap.in'
        with open(str(leap_file), 'w') as outfile:
            outfile.write(tleap_complex)
            
        tleap = f'tleap -f {leap_file}'
        os.system(tleap)

    def prep_protein(self) -> None:
        raw_pdb = self.path / self.pdb
        prep_pdb = self.build_dir / 'prepped.pdb'
        self.pdb = self.build_dir / 'protein.pdb'
       
        # complex workflow for modeling missing residues
        self.triage_pdb(raw_pdb, prep_pdb)
        
        # remove hydrogens (-y) and waters (-d) from the input PDB
        pdb4amber = f'pdb4amber -i {prep_pdb} -o {self.pdb} -y -d'
        os.system(pdb4amber)

    def triage_pdb(self, broken_pdb: PathLike, 
                   repaired_pdb: PathLike) -> str:
        """
        Runs PDBFixer to repair missing loops and ensure structure is
        in good shape. Runs a check against the sequence provided by
        PLINDER and ensures that any non-canonical residues are represented
        in the sequence properly.
        """
        fixer = PDBFixer(filename=str(broken_pdb))
        chains = [chain for chain in fixer.topology.chains()]
        chain_map = {chain.id: [res for res in chain.residues()] 
                     for chain in chains}
        # non-databank models do not have SEQRES and therefore no
        # sequence data to model missing residues
        fixer.sequences = self.inject_fasta(chain_map)
        fixer.findMissingResidues()
        fixer.findMissingAtoms()
        fixer.addMissingAtoms()
        
        with open(str(repaired_pdb), 'w') as f:
            PDBFile.writeFile(fixer.topology, fixer.positions, f)

    def inject_fasta(self, 
                     chain_map: Dict[str, List[str]]) -> Sequences:
        """
        Checks fasta against actual sequence. Modifies sequence so that 
        it correctly matches in the case of non-canonical residues such
        as phosphorylations (i.e. SER -> SEP).
        """
        fasta = open(self.fasta).readlines()
        sequences = []
        for i in range(len(fasta) // 2):
            chain = fasta[2*i].strip().split('.')[-1]
            one_letter_seq = fasta[2*i+1].strip()
            three_letter_seq = [convert_aa_code(aa) for aa in one_letter_seq]
            try:
                three_letter_seq = self.check_ptms(three_letter_seq,
                                                   chain_map[chain])
                sequences.append(
                    Sequence(chainId=chain, 
                             residues=three_letter_seq)
                )
            except KeyError: # not sure what to do if this fails
                break

        return sequences

    def check_ptms(self, 
                   sequence: List[str],
                   chain_residues: List[str]) -> List[str]:
        """
        Check the full sequence (from fasta) against the potentially partial
        sequence from the structural model stored in `chain_residues`.
        """
        for residue in chain_residues:
            resID = int(residue.id) - 1 # since 0-indexed in list
            
            if sequence[resID] != residue.name:
                sequence[resID] = residue.name
        
        return sequence

    def check_ligand(self, ligand: PathLike) -> bool:
        """
        Check ligand for ions and other weird stuff.
        """
        ion = False
        mol = Chem.SDMolSupplier(str(ligand))[0]
        
        ligand = []
        for atom, position in zip(mol.GetAtoms(), mol.GetConformer().GetPositions()):
            symbol = atom.GetSymbol()
            if symbol.lower() in self.cation_list + self.anion_list:
                ion = True

                charge = atom.GetFormalCharge()
                sign = '+' if charge > 1 else '-'
                if abs(charge) > 1:
                    sign = f'{charge}{sign}'

                ligand.append([symbol, sign] + [x for x in position])

        if ion:
            try:
                self.ions.append(ligand)
            except AttributeError:
                self.ions = [ligand]
            return False

        return True

    @property
    def cation_list(self) -> List[str]:
        return [
            'na', 'k', 'ca', 'mn', 'mg', 'li', 'rb', 'cs', 'cu',
            'ag', 'au', 'ti', 'be', 'sr', 'ba', 'ra', 'v', 'cr',
            'fe', 'co', 'zn', 'ni', 'pd', 'cd', 'sn', 'pt', 'hg',
            'pb', 'al'
        ]

    @property
    def anion_list(self) -> List[str]:
        return [
            'cl', 'br', 'i', 'f'
        ]
  
class LigandBuilder:
    def __init__(self, path: PathLike, lig: str, lig_number: int=0):
        self.path = path
        self.lig = path / lig 
        self.ln = lig_number

    def parameterize_ligand(self) -> None:
        """
        Ensures consistent treatment of all ligand sdf files, generating
        GAFF2 parameters in the form of .frcmod and .lib files. Produces
        a mol2 file for coordinates and connectivity and ensures that
        antechamber did not fail. Hydrogens are added in rdkit which 
        generally does a good job of this.
        """
        if self.lig.name[-4:] == '.sdf':
            self.lig = str(self.lig)[:-4]

        fix_resname = f'sed -i "" "s/UNL/LG{self.ln}/" {self.lig}.pdb'
        cleanse_pdb = f'pdb4amber -i {self.lig}.pdb -o {self.lig}_new.pdb'
        convert_to_gaff = f'antechamber -i {self.lig}_new.pdb -fi pdb -o \
                {self.lig}.mol2 -fo mol2 -at gaff2 -c bcc -s 0 -pf y'
        parmchk2 = f'parmchk2 -i {self.lig}.mol2 -f mol2 -o {self.lig}.frcmod'
        
        tleap_ligand = f"""source leaprc.gaff2
        LG{self.ln} = loadmol2 {self.lig}.mol2
        loadamberparams {self.lig}.frcmod
        saveoff LG{self.ln} {self.lig}.lib
        quit
        """
        
        self.add_hydrogens()
        os.system(fix_resname)
        os.system(cleanse_pdb)
        os.system(convert_to_gaff)
        self.move_antechamber_outputs()
        if self.check_sqm():
            os.system(parmchk2)
            leap_file = self.write_leap(tleap_ligand)
            os.system(f'tleap -f {leap_file}')
        else:
            raise RuntimeError('Antechamber has failed! Please take a look at sqm.out')
    
    def add_hydrogens(self) -> None:
        """
        Add hydrogens in rdkit. Atom hybridization is taken from the
        input sdf file and if this is incorrect, hydrogens will be wrong
        too.
        """
        mol = Chem.SDMolSupplier(f'{self.lig}.sdf')[0]
        molH = Chem.AddHs(mol, addCoords=True)
        Chem.MolToPDBFile(molH, f'{self.lig}.pdb')
        
    def move_antechamber_outputs(self) -> None:
        """
        Remove unneccessary outputs from antechamber. Keep the
        sqm.out file as proof that antechamber did not fail.
        """
        os.remove('sqm.in')
        os.remove('sqm.pdb')
        shutil.move('sqm.out', f'{self.lig}_sqm.out')
        
    def check_sqm(self) -> int:
        """
        Checks for evidence that antechamber calculations exited
        successfully. This is always on the second to last line,
        and if not present, indicates that we failed to produce
        sane parameters for this molecule. In that case, I wish
        you good luck.
        """
        line = open(f'{self.lig}_sqm.out').readlines()[-2]
        if 'Calculation Completed' in line:
            return 1
        return 0

    def write_leap(self, inp: str) -> str:
        """
        Writes out a tleap input file and returns the path
        to the file.
        """
        leap_file = f'{self.path}/tleap.in'
        with open(leap_file, 'w') as outfile:
            outfile.write(inp)
            
        return leap_file
    

class ComplexBuilder(ExplicitSolvent):
    """
    Builds complexes consisting of a biomolecule pdb and small molecule ligand.
    Runs antechamber workflow to generate gaff2 parameters.
    """
    def __init__(self, path: str, pdb: str, lig: str, padding: float=10., **kwargs):
        super().__init__(self, path, pdb, padding, **kwargs)
        
        self.lig = lig
        self.parameterize_ligand()
        
    def parameterize_ligand(self):
        """
        Ensures consistent treatment of all ligand sdf files, generating
        GAFF2 parameters in the form of .frcmod and .lib files. Produces
        a mol2 file for coordinates and connectivity and ensures that
        antechamber did not fail. Hydrogens are added in rdkit which 
        generally does a good job of this.
        """
        fix_resname = f'sed -i s/UNL/LIG/ {self.lig}.pdb'
        cleanse_pdb = f'pdb4amber -i {self.lig}.pdb -o {self.lig}_new.pdb'
        convert_to_gaff = f'antechamber -i {self.lig}_new.pdb -fi pdb -o \
                {self.lig}.mol2 -fo mol2 -at gaff2 -c bcc -s 0 -pf y'
        parmchk2 = f'parmchk2 -i {self.lig}.mol2 -f mol2 -o {self.lig}.frcmod'
        
        tleap_ligand = f"""source leaprc.gaff2
        LIG = loadmol2 {self.lig}.mol2
        loadamberparams {self.lig}.frcmod
        saveoff LIG {self.lig}.lib
        quit
        """
        
        self.add_hydrogens()
        os.system(fix_resname)
        os.system(cleanse_pdb)
        os.system(convert_to_gaff)
        self.move_antechamber_outputs()
        if self.check_sqm():
            os.system(parmchk2)
            leap_file = self.write_leap(tleap_ligand)
            os.system(f'tleap -f {leap_file}')
        else:
            raise RuntimeError('Antechamber has failed! Please take a look at sqm.out')
    
    def add_hydrogens(self) -> None:
        """
        Add hydrogens in rdkit. Atom hybridization is taken from the
        input sdf file and if this is incorrect, hydrogens will be wrong
        too.
        """
        mol = Chem.SDMolSupplier(f'{self.lig}.sdf')[0]
        molH = Chem.AddHs(mol, addCoords=True)
        Chem.MolToPDBFile(molH, f'{self.lig}.pdb')
        
    def move_antechamber_outputs(self) -> None:
        """
        Remove unneccessary outputs from antechamber. Keep the
        sqm.out file as proof that antechamber did not fail.
        """
        os.remove('sqm.in')
        os.remove('sqm.pdb')
        shutil.move('sqm.out', f'{self.lig}_sqm.out')
        
    def check_sqm(self) -> int:
        """
        Checks for evidence that antechamber calculations exited
        successfully. This is always on the second to last line,
        and if not present, indicates that we failed to produce
        sane parameters for this molecule. In that case, I wish
        you good luck.
        """
        line = open(f'{self.lig}_sqm.out').readlines()[-2]
        if 'Calculation Completed' in line:
            return 1
        return 0
    
    def assemble_system(self, dim, num_ions) -> None:
        """
        Slightly modified from the parent class, now we have to add
        the ligand parameters and assemble a complex rather than just
        placing a biomolecule in the water box.
        """
        tleap_ffs = '\n'.join([f'source {ff}' for ff in self.ffs])
        tleap_complex = f"""{tleap_ffs}
        source leaprc.gaff2
        loadamberparams {self.lig}.frcmod
        loadoff {self.lig}.lib
        PROT = loadpdb {self.pdb}
        LIG = loadmol2 {self.lig}.mol2
        
        COMPLEX = combine {{PROT LIG}}
        
        setbox COMPLEX centers
        set COMPLEX box {{{dim} {dim} {dim}}}
        solvatebox COMPLEX {self.water_box} {{0 0 0}}
        
        addions COMPLEX Na+ 0
        addions COMPLEX Cl- 0
        
        addIonsRand COMPLEX Na+ {num_ions} Cl- {num_ions}
        
        saveamberparm COMPLEX {self.out}.prmtop {self.out}.inpcrd
        """
        
        leap_file = self.write_leap(tleap_complex)
        tleap = f'tleap -f {leap_file}'
        os.system(tleap)
        



#!/usr/bin/env python
from glob import glob
import MDAnalysis as mda
from natsort import natsorted
import os
from .analysis.analyzer import SimulationAnalyzer
from tqdm import tqdm

path = '/eagle/projects/FoundEpidem/avasan/IDEAL/MiniBinderSims/complexes/sims/tasks'
sim_list = natsorted(list(glob(f'{path}/*')))

for syst in tqdm(sim_list):
    p, s = os.path.split(syst)
    A = SimulationAnalyzer(p, s)
    if isinstance(A.u, mda.Universe) and A.length > 1:
        A.analyze()



from abc import ABC, abstractmethod
from openmm import *
from openmm.app import *
from openmm.unit import *
import mdtraj as md
import numpy as np
import parmed as pmd
import gc
from tqdm import tqdm
from typing import Union

class InteractionEnergy(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def compute(self):
        pass

    @abstractmethod
    def energy(self):
        pass

    @abstractmethod
    def get_selection(self):
        pass

class StaticInteractionEnergy:
    """
    Computes the linear interaction energy between specified chain and other simulation
    components. Can specify a range of residues in chain to limit calculation to. Works on
    a static model but can be adapted to run on dynamics data.

    Inputs:
        pdb (str): Path to input PDB file
        chain (str): Defaults to A. The chain for which to compute the energy between.
            Computes energy between this chain and all other components in PDB file.
        first_residue (int, None): If set, will restrict the calculation to residues
            beginning with resid `first_residue`.
        last_residue (int, None): If set, will restrict the calculation to residues
            ending with resid `last_residue`.
    """
    def __init__(self, pdb: str, chain: str='A', platform: str='CUDA',
                 first_residue: Union[int, None]=None, 
                 last_residue: Union[int, None]=None):
        self.pdb = pdb
        self.chain = chain
        self.platform = Platform.getPlatformByName(platform)
        self.first = first_residue
        self.last = last_residue
        
    def get_system(self) -> None:
        pdb = PDBFile(self.pdb)
        forcefield = ForceField('amber14-all.xml', 'implicit/gbn2.xml')
        system = forcefield.createSystem(pdb.topology,
                                         soluteDielectric=1.,
                                         solventDielectric=80.)

        self.positions = pdb.positions
        self.get_selection(pdb.topology)

        return system

    def compute(self, positions: Union[np.ndarray, None]=None) -> None:
        self.lj = None
        self.coulomb = None

        system = self.get_system()
        if positions is None:
            positions = self.positions
            
        for force in system.getForces():
            if isinstance(force, NonbondedForce):
                force.setForceGroup(0)
                force.addGlobalParameter("solute_coulomb_scale", 1)
                force.addGlobalParameter("solute_lj_scale", 1)
                force.addGlobalParameter("solvent_coulomb_scale", 1)
                force.addGlobalParameter("solvent_lj_scale", 1)

                for i in range(force.getNumParticles()):
                    charge, sigma, epsilon = force.getParticleParameters(i)
                    force.setParticleParameters(i, 0, 0, 0)
                    if i in self.selection:
                        force.addParticleParameterOffset("solute_coulomb_scale", i, charge, 0, 0)
                        force.addParticleParameterOffset("solute_lj_scale", i, 0, sigma, epsilon)
                    else:
                        force.addParticleParameterOffset("solvent_coulomb_scale", i, charge, 0, 0)
                        force.addParticleParameterOffset("solvent_lj_scale", i, 0, sigma, epsilon)

                for i in range(force.getNumExceptions()):
                    p1, p2, chargeProd, sigma, epsilon = force.getExceptionParameters(i)
                    force.setExceptionParameters(i, p1, p2, 0, 0, 0)

            else:
                force.setForceGroup(2)
        
        integrator = VerletIntegrator(0.001*picosecond)

        context = Context(system, integrator, self.platform)
        context.setPositions(positions)
        
        total_coulomb = self.energy(context, 1, 0, 1, 0)
        solute_coulomb = self.energy(context, 1, 0, 0, 0)
        solvent_coulomb = self.energy(context, 0, 0, 1, 0)
        total_lj = self.energy(context, 0, 1, 0, 1)
        solute_lj = self.energy(context, 0, 1, 0, 0)
        solvent_lj = self.energy(context, 0, 0, 0, 1)
        
        coul_final = total_coulomb - solute_coulomb - solvent_coulomb
        lj_final = total_lj - solute_lj - solvent_lj

        self.coulomb = coul_final.value_in_unit(kilocalories_per_mole)
        self.lj = lj_final.value_in_unit(kilocalories_per_mole)
    
    @staticmethod
    def energy(context, solute_coulomb_scale: int=0, solute_lj_scale: int=0, 
               solvent_coulomb_scale: int=0, 
               solvent_lj_scale: int=0) -> float:
        context.setParameter("solute_coulomb_scale", solute_coulomb_scale)
        context.setParameter("solute_lj_scale", solute_lj_scale)
        context.setParameter("solvent_coulomb_scale", solvent_coulomb_scale)
        context.setParameter("solvent_lj_scale", solvent_lj_scale)
        return context.getState(getEnergy=True, groups={0}).getPotentialEnergy()
    
    def get_selection(self, topology) -> None:
        if self.first is None and self.last is None:
            selection = [a.index 
                        for a in topology.atoms() 
                        if a.residue.chain.id == self.chain]
        elif self.first is not None and self.last is None:
            selection = [a.index
                        for a in topology.atoms()
                        if a.residue.chain.id == self.chain 
                        and self.first <= a.residue.id]
        elif self.first is None:
            selection = [a.index
                        for a in topology.atoms()
                        if a.residue.chain.id == self.chain 
                        and self.last >= a.residue.id]
        else:
            selection = [a.index
                        for a in topology.atoms()
                        if a.residue.chain.id == self.sel_chain 
                        and self.first <= a.residue.id <= self.last]

        self.selection = selection

class InteractionEnergyFrame(StaticInteractionEnergy):
    def __init__(self, system: System, top: Topology, 
                 chain: str='A', platform: str='CUDA',
                 first_residue: Union[int, None]=None, 
                 last_residue: Union[int, None]=None):
        super().__init__('', chain, platform, first_residue, last_residue)
        self.system = system
        self.top = top

    def get_system(self):
        self.get_selection(self.top)
        return self.system


class DynamicInteractionEnergy:
    def __init__(self, top: str, traj: str, stride: int=1, 
                 chain: str='A', platform: str='CUDA',
                 first_residue: Union[int, None]=None,
                 last_residue: Union[int, None]=None,
                 progress_bar: bool=False):
        self.system = self.build_system(top, traj)
        self.coordinates = self.load_traj(top, traj)
        self.stride = stride
        self.progress = progress_bar

        self.IE = InteractionEnergyFrame(self.system, self.top, chain, 
                                         platform, first_residue, last_residue)

    def compute_energies(self) -> None:
        n_frames = self.coordinates.shape[0] // self.stride
        self.energies = np.zeros((n_frames, 2))
        
        if self.progress:
            pbar = tqdm(total=n_frames, position=0, leave=False)

        for i in range(n_frames):
            fr = i * self.stride
            self.IE.compute(self.coordinates[fr, :, :])
            self.energies[i, 0] = self.IE.lj
            self.energies[i, 1] = self.IE.coulomb

            if self.progress:
                pbar.update(1)

        if self.progress:
            pbar.close()
    
    def build_system(self, top: str, traj: str) -> System:
        if top[-3:] == 'pdb':
            top = PDBFile(top).topology
            self.top = top
            forcefield = ForceField('amber14-all.xml', 'implicit/gbn2.xml')
            return forcefield.createSystem(top, 
                                           soluteDielectric=1., 
                                           solventDielectric=80.)
        elif top[-6:] == 'prmtop':
            top = AmberPrmtopFile(top)
            self.top = top
            return top.createSystem(nonbondedMethod=CutoffNonPeriodic,
                                    nonbondedCutoff=2. * nanometers,
                                    constraints=HBonds)
        else:
            raise NotImplementedError(f'Error! Topology type {top} not implemented!')

    def load_traj(self, top: str, traj: str) -> np.ndarray:
        return md.load(traj, top=top).xyz

    def setup_pbar(self) -> None:
        self.pbar = tqdm(total=self.coordinates.shape[0], position=0, leave=False)

class DynamicPotentialEnergy:
    """
    Class to compute the interaction energy from MD simulation using OpenMM.
    Inspired by: https://github.com/openmm/openmm/issues/3425
    """
    def __init__(self, top: str, traj: str, seltext: str='protein'):
        self.top = top
        self.traj = traj
        self.selection = seltext

        if top[-3:] == 'pdb':
            self.build_fake_topology(prmtop=False)
        elif top[-6:] == 'prmtop':
            self.build_fake_topology()
        else:
            raise NotImplementedError(f'Error! Topology type {top} not implemented!')
    
    def build_fake_topology(self, prmtop: bool=True) -> None:
        if prmtop:
            top = AmberPrmtopFile(self.top)
            system = top.createSystem(nonbondedMethod=CutoffNonPeriodic,
                                      nonbondedCutoff=2. * nanometers,
                                      constraints=HBonds)
        else:
            top = PDBFile(self.top).topology
            forcefield = ForceField('amber14-all.xml', 'implicit/gbn2.xml')
            system = forcefield.createSystem(top,
                                             soluteDielectric=1.,
                                             solventDielectric=80.)
        
        # Load topology and subset
        topology = md.Topology.from_openmm(top)
        self.sub_ind = topology.select(self.selection)
        
        sub_top = topology.subset(self.sub_ind)
        self.old_topology = topology
        self.topology = sub_top.to_openmm()
        
        # Create protein only system
        structure = pmd.openmm.load_topology(top, system)
        sturcture = structure[self.sub_ind]

        # Add HBond restraints if in explicit water?
        if prmtop:
            new_bond_type = pmd.topologyobjects.BondType(k=400, req=1.)
            constrained_bond_type = structure.bond_types.append(new_bond_type)
            structure.bond_types.claim()

            for bond in structure.bonds:
                if bond.type is None:
                    bond.type = new_bond_type

            # Create new system
            new_system = structure.createSystem(nonbondedMethod=CutoffNonPeriodic, 
                                                nonbondedCutoff=2. * nanometers)

        else:
            new_system = structure.createSystem(soluteDielectric=1.,
                                                solventDielectric=80.)

        self.system = new_system
        integrator = LangevinMiddleIntegrator(300*kelvin, 1/picosecond, 0.004 * picoseconds)
        self.simulation = Simulation(self.topology, self.system, integrator)
    
    def compute(self) -> None:
        full_traj = md.load(self.traj, top=self.top)
        self.energies = np.zeros((full_traj.n_frames))
        for fr in range(full_traj.n_frames):
            energy = self.calc_energy(full_traj.xyz[fr, self.sub_ind, :])
            self.energies[fr] = energy._value

    def calc_energy(self, positions) -> float:
        self.simulation.context.setPositions(positions)
        state = self.simulation.context.getState(getEnergy=True)

        return state.getPotentialEnergy()

class PairwiseInteractionEnergy(StaticInteractionEnergy):
    """
    Computes the pairwise interaction energy between a single residue from one 
    selection and the entirety of another selection.
    """
    def __init__(self, system: System, top: Topology, 
                 mda_sel1: str, mda_sel2: str,
                 chain: str='A', platform: str='CUDA'):
        super().__init__('', chain, platform, None, None)
        self.system = system
        self.top = top

    def get_system(self):
        self.get_selection(self.top)
        return self.system

    def get_selection(self):
        self.selection = pass



#!/usr/bin/env python
from glob import glob
from math import ceil
import parsl
from molecular_simulations.simulate import parsl_config
import os

# Housekeeping stuff
run_dir = '/eagle/projects/FoundEpidem/msinclair/tools/molecular-dynamics/examples'

paths = sorted([f for f in glob('test_sims/*')])
n_jobs = len(paths)
gpus_per_node = 4 # do not change this; 4 per node on Polaris
n_nodes = n_jobs // gpus_per_node # not worth the risk of `ceil`ing a bad flop (e.g. 1.00001)

worker_init_cmd = f'module use /soft/modulefiles/; module load conda; conda activate simulate; \
        cd {run_dir}'

# PBS options for Parsl
user_opts = {
        'worker_init': worker_init_cmd,
        'scheduler_options': '#PBS -l filesystems=home:eagle',
        'account': 'FoundEpidem',
        'queue': 'preemptable',
        'walltime': '72:00:00',
        'nodes_per_block': n_nodes,
        'cpus_per_node': 32,
        'available_accelerators': 4,
        'cores_per_worker': 8
        }

debug_opts = {
        'worker_init': worker_init_cmd,
        'scheduler_options': '#PBS -l filesystems=home:eagle',
        'account': 'FoundEpidem',
        'queue': 'debug',
        'walltime': '0:30:00',
        'nodes_per_block': n_nodes,
        'cpus_per_node': 32,
        'available_accelerators': 4,
        'cores_per_worker': 8
        }

cfg = parsl_config.get_config(run_dir, user_opts)

sim_length = 10 # ns
timestep = 4 # fs
n_steps = sim_length / timestep * 1000000

@parsl.python_app
def run_md(path: str, eq_steps=500_000, steps=250_000_000):
    from molecular_simulations.simulate.omm_simulator import Simulator

    simulator = Simulator(path, equil_steps=eq_steps, prod_steps=steps)
    simulator.run()

parsl_cfg = parsl.load(cfg)

futures = []
for path in paths:
    futures.append(run_md(path, steps=n_steps))

outputs = [x.result() for x in futures]



from molecular_simulations import AuroraSettings
from parsl import get_config, python_app
from pathlib import Path
from plinder.core.scores import query_index

input_dir = Path('/lus/flare/projects/FoundEpidem/plinder/2024-06/v2/systems') 
root_simulation_dir = Path('/lus/flare/projects/FoundEpidem/plinder/simulations')

columns_of_interest = [
    'system_id', 'entry_pdb_id', 'entry_oligomeric_state', 'entry_resolution',
    'system_num_ligand_chains', 'system_protein_chains_num_unresolved_residues',
    'ligand_is_artifact', 'system_protein_chains_validation_unknown_residue_count',
    'system_ligand_validation_unknown_residue_count', 'system_num_covalent_ligands',
]

filters = [
    ('system_num_ligand_chains', '>=', 1),
    ('ligand_is_artifact', '==', False),
    ('system_ligand_validation_unknown_residue_count', '==', 0.),
    ('system_num_covalent_ligands', '==', 0),
]

df = query_index(columns=columns_of_interest, filters=filters)
df.drop_duplicates(subset='system_id', inplace=True)

@python_app
def construct(system_id, input_path, output_path):
    molecular_simulations.build.build_amber import LigandError, PLINDERBuilder
    import os
    from pathlib import Path
    
    path = path / system_id
    out_path = root_simulation_dir / system_id
    # check if we have already downloaded this system
    if not os.path.exists(path):
        builder = PLINDERBuilder(input_path, 
                                 system_id=system_id, 
                                 out=output_path)
        try:
            builder.build()
        except LigandError:
            return system_id

future = []
for system_id in df['system_id']:
    futures.append(construct(system_id, input_path, output_path))

bad_systems = [x.result() for x in futures]

with open('bad_systems.txt') as out:
    for system in bad_systems:
        out.write(f'{system}\n')



#!/usr/bin/env python
from abc import ABC, abstractmethod
from dataclasses import dataclass
from parsl.config import Config
from parsl.providers import PBSProProvider
from parsl.executors import HighThroughputExecutor
from parsl.launchers import MpiExecLauncher, GnuParallelLauncher
from parsl.addresses import address_by_interface
from parsl.utils import get_all_checkpoints
from pathlib import Path
from typing import Union

PathLike = Union[str, Path]

@dataclass
class BaseComputeSettings(ABC):
    """Compute settings (HPC platform, number of GPUs, etc)."""

    @abstractmethod
    def config_factory(self, run_dir: PathLike) -> Config:
        """
        Create new Parsl configuration.
        """

class PolarisSettings(BaseComputeSettings):
    num_nodes: int = 1
    worker_init: str = ''
    scheduler_options: str = ''
    account: str
    queue: str
    walltime: str
    cpus_per_node: int = 64
    strategy: str = 'simple'

    def config_factory(self, run_dir: PathLike) -> Config:
        """Create a configuration suitable for running all tasks on single nodes of Polaris
        We will launch 4 workers per node, each pinned to a different GPU
        Args:
            num_nodes: Number of nodes to use for the MPI parallel tasks
            user_options: Options for which account to use, location of environment files, etc
            run_dir: Directory in which to store Parsl run files. Default: `runinfo`
        """
        return Config(
            retries=1,  # Allows restarts if jobs are killed by the end of a job
            executors=[
                HighThroughputExecutor(
                    label=self.label,
                    heartbeat_period=15,
                    heartbeat_threshold=120,
                    worker_debug=True,
                    available_accelerators=4,  # Ensures one worker per accelerator
                    address=address_by_hostname(),
                    cpu_affinity="alternating",
                    prefetch_capacity=0,  # Increase if you have many more tasks than workers
                    start_method="spawn",
                    provider=PBSProProvider(  # type: ignore[no-untyped-call]
                        launcher=MpiExecLauncher(
                            bind_cmd="--cpu-bind", overrides="--depth=64 --ppn 1"
                        ),  # Updates to the mpiexec command
                        account=self.account,
                        queue=self.queue,
                        select_options="ngpus=4",
                        # PBS directives (header lines): for array jobs pass '-J' option
                        scheduler_options=self.scheduler_options,
                        worker_init=self.worker_init,
                        nodes_per_block=self.num_nodes,
                        init_blocks=1,
                        min_blocks=0,
                        max_blocks=1,  # Can increase more to have more parallel jobs
                        cpus_per_node=self.cpus_per_node,
                        walltime=self.walltime,
                    ),
                ),
            ],
            run_dir=str(run_dir),
            strategy=self.strategy,
            app_cache=True,
        )

class AuroraSettings(BaseComputeSettings):
    worker_init: str = ""
    num_nodes: int = 1
    scheduler_options: str = ""
    account: str
    queue: str
    walltime: str
    retries: int = 0
    cpus_per_node: int = 48 # only 4 cpus per OpenMM job
    strategy: str = "simple"

    def config_factory(self, run_dir: PathLike) -> Config:
        """Create a Parsl configuration for running on Aurora."""
        accel_ids = [str(i) for i in range(12)]
        return Config(
            executors=[
                HighThroughputExecutor(
                    label=self.label,
                    available_accelerators=accel_ids,  # Ensures one worker per accelerator
                    cpu_affinity="block",  # Assigns cpus in sequential order
                    prefetch_capacity=0,
                    max_workers=12,
                    cores_per_worker=16,
                    heartbeat_period=30,
                    heartbeat_threshold=300,
                    worker_debug=False,
                    provider=PBSProProvider(
                        launcher=MpiExecLauncher(
                            bind_cmd="--cpu-bind",
                            overrides=f"--depth=208 --ppn 1"
                        ),  # Ensures 1 manger per node and allows it to divide work among all 208 threads
                        worker_init=self.worker_init,
                        nodes_per_block=self.num_nodes,
                        account=self.account,
                        queue=self.queue,
                        walltime=self.walltime,

                    ),
                ),
            ],
            run_dir=str(run_dir),
            checkpoint_mode='task_exit',
            retries=self.retries,
            app_cache=True,
        )


@dataclass
class AuroraCPUSettings:
    worker_init: str = ""
    num_nodes: int = 1
    scheduler_options: str = ""
    account: str
    queue: str
    walltime: str
    retries: int = 0
    cpus_per_node: int = 208
    strategy: str = "simple"

    def config_factory(self, run_dir: PathLike) -> Config:
        """Create a Parsl configuration for running on CPU only on Aurora."""
        return Config(
            executors=[
                HighThroughputExecutor(
                    label='Aurora_CPU',
                    available_accelerators=None,
                    cpu_affinity="block",  # Assigns cpus in sequential order
                    prefetch_capacity=0,
                    max_workers=208,
                    cores_per_worker=1,
                    heartbeat_period=30,
                    heartbeat_threshold=300,
                    worker_debug=False,
                    provider=PBSProProvider(
                        launcher=MpiExecLauncher(
                            bind_cmd="--cpu-bind",
                            overrides=f"--depth=208 --ppn 1"
                        ),  # Ensures 1 manger per node and allows it to divide work among all 208 threads
                        worker_init=self.worker_init,
                        nodes_per_block=self.num_nodes,
                        account=self.account,
                        queue=self.queue,
                        walltime=self.walltime,

                    ),
                ),
            ],
            run_dir=str(run_dir),
            checkpoint_mode='task_exit',
            retries=self.retries,
            app_cache=True,
        )



#!/usr/bin/env python
from .analysis.funcs import (CoMDist, ContactFrequency, 
                             DeltaCOM, RadiusofGyration)
import gc
import MDAnalysis as mda
from MDAnalysis.analysis import align
from MDAnalysis.analysis.contacts import Contacts
from MDAnalysis.analysis.distances import distance_array
from MDAnalysis.analysis.rms import RMSD
from MDAnalysis.analysis.base import AnalysisBase
import numpy as np
import os
from typing import Union

class SimulationAnalyzer:
    def __init__(self, path: str, system: str, data_path: str='sim_data', 
                 stride: int=100):
        self.path = path
        self.system = system
        self.data_path = data_path
        self.stride = stride
        self.u = self.build_universe()

    def build_universe(self):
        pdb = f'{self.path}/{self.system}/protein.pdb'
        dcd = f'{self.path}/{self.system}/sim.dcd'
        try:
            u = mda.Universe(pdb, dcd, in_memory=True, in_memory_step=self.stride)
        except OSError:
            return 0

        return u

    def analyze(self):
        analyses = [self.binderRMSD, self.nativeContacts, self.deltaCoM,
                    self.comDist, self.rog, self.contactProb]
        dirs = ['rmsd_ppi', 'contacts_ppi', 'com_ppi', 
                'cdist_ppi', 'rog_ppi', 'cont_prob_ppi']
        sufs = [None, None, 'coms', 
                'raw', None, 'raw']

        for a, d, s in zip(analyses, dirs, sufs):
            if not os.path.exists(self.get_output(d, suffix=s)):
                a()

    def binderRMSD(self):
        R = RMSD(self.u, self.u, select='chainID A and backbone', 
                 groupselections=['chainID B and backbone'])
        R.run()

        np.save(self.get_output('rmsd_ppi'), R.rmsd[:, 3])

    def nativeContacts(self):
        sel_text = 'segid A and not type H'
        ref_text = 'segid B and not type H'
        sel = self.u.select_atoms(sel_text)
        ref = self.u.select_atoms(ref_text)

        ca1 = Contacts(self.u, select=(sel_text, ref_text),
                       refgroup=(sel, ref), radius=5, 
                       method='soft_cut')
        ca1.run()

        np.save(self.get_output('contacts_ppi'), ca1.timeseries)

    def deltaCoM(self):
        ref = self.u.select_atoms('segid B and not type H')

        dc = DeltaCOM(ref, residence=False)
        dc.run()

        np.save(self.get_output('com_ppi', suffix='coms'), dc.coms)

    def comDist(self):
        sel = self.u.select_atoms('segid A and backbone')
        ref = self.u.select_atoms('segid B and backbone')

        cd = CoMDist(sel, ref)
        cd.run()

        np.save(self.get_output('cdist_ppi', suffix='raw'), cd.com_dists)

    def rog(self):
        sel = self.u.select_atoms('segid B and name CA')
        
        rg = RadiusofGyration(sel)
        rg.run()

        np.save(self.get_output('rog_ppi'), rg.results)

    def contactProb(self):
        binder = self.u.select_atoms('segid B and name CA')
        target = self.u.select_atoms('segid A and name CA')

        cf = ContactFrequency(binder, target)
        cf.run()

        np.save(self.get_output('cont_prob_ppi', suffix='raw'), cf.contacts)
        np.save(self.get_output('cont_prob_ppi', suffix='prob'), cf.contact_probabilities)

    def get_output(self, d, suffix: Union[None, str]=None):
        if suffix is not None:
            ext = f'{suffix}.npy'
        else:
            ext = 'npy'

        return f'{self.data_path}/{d}/{self.system}.npy'

    def clean(self):
        del self.u
        gc.collect()

    @property
    def length(self):
        return len(self.u.trajectory)



#!/usr/bin/env python
from molecular_simulations.build.build_amber import ExplicitSolvent

path = '.'

for m in range(4):
    mpath = f'{path}/test_sim/model{m}'
    pdb = f'{path}/test_build/chai-model.pdb'

    builder = ExplicitSolvent(mpath, pdb, protein=True)
    builder.build()
